/***************/
/* "deconv4.c" */
/***************/

/***** INTERFACE *****/

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "typedef4.h"	/* InstrParam is defined */

#undef DEBUG
#define TRY_EXTRAPOLATION
#define REPORT
#define TEMP_FILE "temp.txt"

/*** Global Variables (external) ***/
extern unsigned long nData;		/* Number of source data; declared in "main4.c" */
extern double *twot;			/* 2theta data; declared in "main4.c" */
extern double *scat;			/* Intensity data; declared in "main4.c" */
extern double *err;			/* Error data; declared in "main4.c" */
extern double *err2;			/* Error data, type2; declared in "main4.c" */
extern InstrParam instr;		/* Instrumental parameters */
extern int EvaluateAnotherErrorIndication;

/*** Function Prototypes (uses) ***/
void error(char error_text[]);
	/* in "error.c" */
void SplineNonEq2Eq(double xS[], double yS[], unsigned long nS, 
	double xStart, double xStep, double yD[], unsigned long nD);
	/* in "spline.c" */
	/* Cubic spline interpolation from non-equidistantly sampled data */
	/* to equidistantly sampled data */
void SplineEq2NonEq(double xStart, double xStep, double yS[], unsigned long nS,
	double xD[], double yD[], unsigned long nD);
	/* in "spline.c" */
	/* Cubic spline interpolation from equidistantly sampled data */
	/* to non-equidistantly sampled data */
void FTKalpha1(double *re,double *im,double x,double g1);
	/* in "FTInstrumental4.c" */
void FTKalpha2(double *re,double *im,double x,double rho,double x2,double dg);
	/* in "FTInstrumental4.c" */
void FTGauss(double *re,double *im,double x,double gG);
	/* in "FTInstrumental4.c" */
void FTUpperAxial(double *re,double *im,double xi,double PhiA);
	/* in "FTInstrumental4.c" */
void FTLowerAxial(double *re,double *im,double xi,double PhiA);
	/* in "FTInstrumental4.c" */
void FTFlatSpecimen(double *re,double *im,double xi,double Phi);
	/* in "FTInstrumental4.c" */
void FTTransparency(double *re,double *im,double xi,double invMuR);
	/* in "FTInstrumental4.c" */
void FTReceivingSlit(double *re, double *im, double xi, double rr);
    /* in "FTInstrumental4.c" */
void fft(double data[], unsigned long n);
	/* in "fft.c" */
	/* Replaces real array of data[0..n-1] by its discrete fourier transform */
	/* The real part of Fourier transform is stored in */
	/* data[0],data[2],...,data[n/2],data[1] (!) ; */
	/* the imaginary is in data[3],data[5],...,data[n/2+1] */
void ifft(double data[], unsigned long n);	/* in "fft.c" */
	/* The inverse transform of fft() */

/*** Function Prototypes (public) ***/
void RemoveAllInstrumental(void);	/* in "deconv4.c" */
void RemoveInstrumental(int flag);	/* in "deconv4.c" */
	/* flag == 1: Kalpha1, Kalpha2 & higer-axial divergence */
	/* flag == 2: lower-axial divergnence */
	/* flag == 3; transparency */
	/* flag == 4; flat-specimen */
	/* flag == 5; X-ray source & receiving slit */

/***** IMPLEMENTATION *****/

#define PI 3.141592653589793238462643383279

/*** Global Variables (private) ***/
static unsigned long nNonZero;	/* Number of interpolated data */
static unsigned long nTotal;	/* Number of interpolated data with zero-padding */
static double x0start,x0step;	/* start point & interval of interpolated abscissa */
static double xistep;			/* interval in Fourier transformed data */
static double *x0T=NULL;		/* scale-transformed abscissa */
static double *y0i=NULL;		/* interpolated intensity data */
static double *s0=NULL;			/* interpolated error data */
static double *s1=NULL;			/* interpolated error data (2) */
static double *ft_Instrum=NULL;	/* Fourier transform of instrumental function */
static double *ft_Instrum2=NULL;	/* Fourier transform of instrumental function (2) */

/*** Function Prototypes (private) ***/
static double sqr(double x);
static double CorrectionFactor(double twotheta);
static double CalcInstrWidth(int flag);
static double SpecificWidth(int flag, double tt);
static double ScaleTransform(int flag, double tt);
static void PrepareParameters(int flag);
static void PreCorrectInterpolatedError(int flag);
static void CreateScaledData(int flag);
static void InterpolateData(void);
static void CreateFTInstrumental(int flag,unsigned long n);
static void ReCorrectInterpolatedError(int flag);
static void DeconvoluteInstrumental(int flag,unsigned long n);
static void PickupData(void);
static void ClearData2(void);

static double sqr(double x)
/* returns square */
{
	return x*x;
}

static double CorrectionFactor(double twotheta)
/* calculates combination of Lorentz(1/sin(theta)), */
/* polarization( (1 + A cos(2theta)^2)/(1 + A) ), */
/* and powder(1/sin(2theta))  factors */
/* twotheta: 2 theta in degree */
{
	double twot;	/* 2 theta in radian */
	double A;
	A=0.799466;	/* A = cos(2thetaM)^2, */
		/* assuming graphite (c=6.696) (002) monochromator */
		/* and CuK-alpha (lambda=1.5406) */
	twot=twotheta*PI/180;
	return (1+A*sqr(cos(twot)))/((1+A)*sin(twot/2)*sin(twot));
}

static double CalcInstrWidth(int flag)
/* calculates rough width of the instrumental functions */
/* used only for determine zero-padding width */
{
	double w;

	if (flag==1) {
		w = fabs(instr.x2)*10 + fabs(instr.g1+instr.dg2)*32;
	} else if (flag==2) {
		w = fabs(sqr(instr.phiA*PI/360))*20;
		/* w += fabs(sqr(instr.phiE*PI/360))*10; */
 	} else if (flag==3) {
		w = fabs(instr.invmuR)*20;
	} else if (flag==4) {
		w = fabs(sqr(instr.phiE*PI/360))*20;
	} else if (flag==5) {
		w = fabs(instr.S2R)*20+fabs(instr.R2R)*20;
	} else {
		w = instr.gGmax*20;
	}
	return w;
}

static double SpecificWidth(int flag, double tt)
/* calculates angular dependent width for "tt" in degree. */
{
	double ans;
	if (flag==1) {
		ans = 360/PI*tan(tt*PI/360);
	} else if (flag==2) {
		ans = 360/PI/tan(tt*PI/360);
	} else if (flag==3) {
		ans = 90*sin(tt*PI/180)/PI;
	} else if (flag==4) {
		if (tt > instr.twotC) {
			ans = 360/PI/tan(tt*PI/360);
		} else {
			ans = 180/PI*sin(tt*PI/180)/sqr(instr.sintC);
		}
	} else if (flag==5) {
		ans = 1;
	} else {
		ans = 1;
	}
	return ans;
}

static double ScaleTransform(int flag, double tt)
/* calculates transformed scale for "tt" in degree. */
{
	double sc2,ans;

	if (flag==1) {
		ans = log(sin(tt*PI/360));
	} else if (flag==2) {
		ans = -log(cos(tt*PI/360));
	} else if (flag==3) {
		ans = 2*log(tan(tt*PI/360));
	} else if (flag==4) {
		sc2=instr.sintC*instr.sintC;
		if (tt > instr.twotC) {
			ans = -log(cos(tt*PI/360))+0.5*(1-sc2)*log(1-sc2)+sc2*log(instr.sintC);
		} else {
			ans = sc2*log(tan(tt*PI/360));
		}
	} else if (flag==5) {
		ans = tt;
	} else {
		ans = tt;
	}
	return ans;
}
		
static void PrepareParameters(int flag)
/* Prepare parameters for interpolation */
///* This routine creates x0T[0..nData+1] */
/* This routine creates x0T[0..nData-1] */
{
	double x0range,fillZeroRange;
	unsigned long i;
	
#ifdef TRY_EXTRAPOLATION
	x0T = malloc((size_t)(nData+2)*sizeof(double));
#else
	x0T = malloc((size_t)nData*sizeof(double));
#endif

	fillZeroRange = CalcInstrWidth(flag);
	for (i=0; i<nData; i++) {
		x0T[i] = ScaleTransform(flag, twot[i]);
	}

#ifdef TRY_EXTRAPOLATION
	x0T[nData] = x0T[nData-1]+0.5*fillZeroRange;
	x0T[nData+1] = x0T[nData-1]+fillZeroRange;//x0T[nData]+x0T[1]-x0T[0];
	scat[nData] = 0.5*(scat[nData-1]+scat[0]);//scat[0];
	scat[nData+1] = scat[0];//scat[1];
	err[nData] = 0.5*(err[nData-1]+err[0]);//err[0];
	err[nData+1] = err[0];//err[1];
	if (EvaluateAnotherErrorIndication == 1) {
		err2[nData] = 0.5*(err2[nData-1]+err2[0]);//err2[0];
		err2[nData+1] = err2[0];//err2[1];
	}
#endif
	x0step=x0T[1]-x0T[0];
	for (i=2; i<nData; i++) {
		if (x0T[i]-x0T[i-1] < x0step) {
			x0step = x0T[i]-x0T[i-1];
		}
	}
	x0range=x0T[nData-1]-x0T[0];
		/* x-Range in the Scaled Source Data */
	nTotal=(unsigned long)ceil((x0range+fillZeroRange)/x0step);
		/* Minimum number required for interpolated data */
	i=(unsigned long)ceil(log(nTotal)/log(2));
	nTotal=(unsigned long)pow(2,i);
		/* Total number required for interpolation */
	nNonZero=(unsigned long)floor(nTotal*x0range/(x0range+fillZeroRange));
		/* Number of non-zero values of interpolated data */
	x0start = x0T[0];
	x0step = (x0T[nData-1]-x0T[0])/(nNonZero-1);
	xistep = 1.0/(nTotal*x0step);

	fprintf(stderr, "Step %d: %d points, including %d zeropadding\n",
		(int)flag,(int)nTotal,(int)(nTotal-nNonZero));
}
	
/*** InterpolateData ***/
static void InterpolateData(void)
/* uses nData,x0T[],scat[],err[],nNonZero,nTotal,x0start,x0step */
/* This routine creates y0i[],s0[]. */
{
	unsigned long nD;
#ifdef DEBUG
	FILE *fp;
	unsigned long i;
#endif
#ifdef TRY_EXTRAPOLATION
	nD=nData+2;
#else
	nD=nData;
#endif;
	/*** Interpolate Intensity Data ***/
	y0i = malloc(nTotal*sizeof(double));
	if (y0i==NULL) {
		error("Could not allocate memories for y0i");
	}
	SplineNonEq2Eq(x0T,scat,nD,x0start,x0step,y0i,nTotal);

	/*** Interpolate Variance Data ***/
	s0 = malloc(nTotal*sizeof(double));
	if (s0==NULL) {
		error("Could not allocate memories for s0");
	}
	SplineNonEq2Eq(x0T,err,nD,x0start,x0step,s0,nTotal);

	if (EvaluateAnotherErrorIndication == 1) {
		/*** Interpolate Variance Data (2) ***/
		s1 = malloc(nTotal*sizeof(double));
		if (s1==NULL) {
			error("Could not allocate memories for s1");
		}
		SplineNonEq2Eq(x0T,err2,nD,x0start,x0step,s1,nTotal);
	}

#ifdef DEBUG
	fp = fopen(TEMP_FILE,"wt");
	for (i=0; i<nTotal; i++) {
		if (EvaluateAnotherErrorIndication == 1) {
			fprintf(fp, "%f %f %f %f\n",x0start+x0step*i,y0i[i],s0[i],s1[i]);
		} else {
			fprintf(fp, "%f %f %f\n",x0start+x0step*i,y0i[i],s0[i]);
		}
	}
	fclose(fp);
#endif
}

/*** Cmul ***/
static void Cmul(double *re3,double *im3,double re1,double im1,double re2, double im2)
/* (*re3,*im3) = (re1,im1) * (re2,im2) */
{
	*re3 = re1*re2-im1*im2;
	*im3 = re1*im2+re2*im1;
}
/*** Cdiv() ***/
static void Cdiv(double *re3,double *im3,double re1,double im1,double re2,double im2)
/* (*re3,*im3) = (re1,im1) / (re2,im2) */
{
	double arg;
	arg = re2*re2+im2*im2;
	*re3 = (re1*re2+im1*im2)/arg;
	*im3 = (re2*im1-re1*im2)/arg;
}

static void CreateFTInstrumental(int flag,unsigned long n)
/* Create Fourier transform of instrumental function */
/* This routine creates ft_Instrum[] (& ft_Instrum2) */
{
	unsigned long i;
	double gG2,xic;
	double re1,im1,re2,im2;

	ft_Instrum = malloc(n*sizeof(double));
	if (ft_Instrum==NULL) {
		error("Could not allocate memories for ft_Instrum");
	}
	if (EvaluateAnotherErrorIndication == 1) {
		ft_Instrum2 = malloc(n*sizeof(double));
		if (ft_Instrum2==NULL) {
			error("Could not allocate memories for ft_Instrum2");
		}
	}

	for (i=0; i<=n/2; i++) {	/* Note i = 0,...,n/2 ! */
		if (flag==1) {
			FTKalpha1(&re1,&im1,i*xistep,instr.g1);
			FTKalpha2(&re2,&im2,i*xistep,instr.rho,instr.x2,instr.dg2);
			Cmul(&re1,&im1,re1,im1,re2,im2);
			FTUpperAxial(&re2,&im2,i*xistep,instr.phiA);
			Cmul(&re1,&im1,re1,im1,re2,im2);
		} else if (flag==2) {
			FTLowerAxial(&re1,&im1,i*xistep,instr.phiA);
		} else if (flag==3) {
			FTTransparency(&re1,&im1,i*xistep,instr.invmuR);
		} else if (flag==4) {
			FTFlatSpecimen(&re1,&im1,i*xistep,instr.phiE);
		} else if (flag==5) {
			FTReceivingSlit(&re1,&im1,i*xistep,instr.S2R);
			FTReceivingSlit(&re2,&im2,i*xistep,instr.R2R);
			Cmul(&re1,&im1,re1,im1,re2,im2);
		} else {
			FTGauss(&re1,&im1,i*xistep,instr.gGmax);
		}
		if (i==0) {
			ft_Instrum[0] = re1;
			if (EvaluateAnotherErrorIndication == 1) {
				ft_Instrum2[0] = 1/re1;
			}
		} else if (i<n/2) {
			ft_Instrum[2*i] = re1;
			ft_Instrum[2*i+1] = im1;
			if (EvaluateAnotherErrorIndication == 1) {
				Cdiv(&re1,&im1,1,0,re1,im1);
				ft_Instrum2[2*i] = re1;
				ft_Instrum2[2*i+1] = im1;
			}
		} else {	/* (i==n/2) */
			ft_Instrum[1] = re1;
			if (EvaluateAnotherErrorIndication == 1) {
				ft_Instrum2[1] = 1/re1;
			}
		}
	}

	if ((1 <= flag) && (flag <= 4)) {
		xic=(n/2-1)*xistep;
		gG2=-log( sqr(ft_Instrum[n-2]) + sqr(ft_Instrum[n-1]) )/( 2 * sqr(PI*xic) );
		instr.gG0 = sqrt(sqr(instr.gG0) + gG2 * sqr(SpecificWidth(flag,twot[0])));
		if (90.0<twot[0]) {
			instr.gG1 = sqrt(sqr(instr.gG1) + gG2 * sqr(SpecificWidth(flag,twot[0])));
		} else if (twot[nData-1]<90.0) {
			instr.gG1 = sqrt(sqr(instr.gG1) + gG2 * sqr(SpecificWidth(flag,90.0)));
		} else {
			instr.gG1 = sqrt(sqr(instr.gG1) + gG2 * sqr(SpecificWidth(flag,twot[nData-1])));
		}
		instr.gG2 = sqrt(sqr(instr.gG2) + gG2 * sqr(SpecificWidth(flag,twot[nData-1])));
	} else {
		instr.gG0 = sqrt(sqr(instr.gG0) + (sqr(instr.S2R)+sqr(instr.R2R))/4/log(2));
		instr.gG1 = sqrt(sqr(instr.gG1) + (sqr(instr.S2R)+sqr(instr.R2R))/4/log(2));
		instr.gG2 = sqrt(sqr(instr.gG2) + (sqr(instr.S2R)+sqr(instr.R2R))/4/log(2));
	}
	instr.gGmax = instr.gG0;
	if (instr.gG1 > instr.gGmax) {
		instr.gGmax = instr.gG1;
		if (instr.gG2 > instr.gGmax) {
			instr.gGmax = instr.gG2;
		}
	}
	/* */
	fprintf(stderr,"specific widths = %f %f %f %f\n",instr.gG0,instr.gG1,instr.gG2,instr.gGmax);
	/* */

}

static void DeconvoluteInstrumental(int flag, unsigned long n)
/* deconvolutes instrumental function ***/
{
	unsigned long i;

	/***********************************/
	/* Deconvolution of intensity data */
	/***********************************/
	fft(y0i,n);	/* Fourier transform of y0i */
	y0i[0] = y0i[0]/ft_Instrum[0];	/* 0th & 1st elements are separately treated */
	y0i[1] = y0i[1]/ft_Instrum[1];
	for (i=1; i<n/2; i++) {	/* Note i = 1,...,(n/2-1) ! */
		Cdiv(&y0i[2*i],&y0i[2*i+1],
			y0i[2*i],y0i[2*i+1],ft_Instrum[2*i],ft_Instrum[2*i+1]);
			/* Division by Fourier transformed instrumental function */
	}
	ifft(y0i,n);	/* inverse Fourier transform of y0i */

	/***********************************/
	/* Deconvolution of error data (1) */
	/***********************************/
	ifft(ft_Instrum,n);	/* inverse Fourier transform of ft_Instrum */
	for (i=0; i<n; i++) {
		ft_Instrum[i] = sqr(ft_Instrum[i]);
	}	/* now, ft_Instrum is squared instrumental function */
	fft(ft_Instrum,n);	/* Fourier transform */
		/* now, ft_Instrum is Fourier transform of squared instrumental function */
	/*** Calculate reciprocal variance ***/
	for (i=0; i<n; i++) {
		if (s0[i]!=0.0) {
			s0[i] = 1/sqr(s0[i]);
		} else {
			s0[i]=0.0;
		}
	}	/* now, s0 is reciprocal variance */
	fft(s0,n);	/* Fourier transform */
	s0[0] = s0[0] * ft_Instrum[0];	/* 0th & 1st elements are separately treated */
	s0[1] = s0[1] * ft_Instrum[1];
	for (i=1; i<n/2; i++) {	/* Note i = 1,...,(n/2-1) ! */
		Cmul(&s0[2*i],&s0[2*i+1],
			s0[2*i],s0[2*i+1],ft_Instrum[2*i],-ft_Instrum[2*i+1]);
			/* Multiplication of conjugate Fourier transform of */
			/* squared instrumental function */
	}
	ifft(s0,n);	/* inverse Fourier transform of s0 */
	/*** Calculate error ***/
	for (i=0; i<n; i++) {
		if (s0[i] == 0.0) {
			s0[i]=0.0;
		} else {
			s0[i] = 1/sqrt(fabs(s0[i]));
		}
	}

	if (EvaluateAnotherErrorIndication==1) {
		/***********************************/
		/* Deconvolution of error data (1) */
		/***********************************/
		ifft(ft_Instrum2,n);	/* inverse Fourier transform of ft_Instrum2 */
		for (i=0; i<n; i++) {
			ft_Instrum2[i] = sqr(ft_Instrum2[i]);
		}	/* now, ft_Instrum2 is squared inverse instrumental function */
		fft(ft_Instrum2,n);	/* Fourier transform */
			/* now, ft_Instrum2 is Fourier transform of */
			/* squared inverse instrumental function */
		/*** Calculate variance ***/
		for (i=0; i<n; i++) {
			s1[i] = sqr(s1[i]);
		}	/* now, s1 is variance */
		fft(s1,n);	/* Fourier transform */
		s1[0] = s1[0] * ft_Instrum2[0];
		s1[1] = s1[1] * ft_Instrum2[1];
		/* 0th & 1st elements are separately treated */
		for (i=1; i<n/2; i++) {	/* Note i = 1,...,(n/2-1) ! */
			Cmul(&s1[2*i],&s1[2*i+1],
				s1[2*i],s1[2*i+1],ft_Instrum2[2*i],ft_Instrum2[2*i+1]);
				/* Multiplication of Fourier transform of */
				/* squared inverse instrumental function */
		}
		ifft(s1,n);	/* inverse Fourier transform of s1 */
		/*** Calculate error ***/
		for (i=0; i<n; i++) {
			if (s1[i] == 0.0) {
				s1[i]=0.0;
			} else {
				s1[i] = sqrt(fabs(s1[i]));
			}
		}
	}
}

static void ConvoluteInstrumental(int flag, unsigned long n)
/* deconvolutes instrumental function ***/
{
	unsigned long i;

	/*** Convolution of intensity data ***/
	fft(y0i,n);	/* Fourier transform of y0i */
	y0i[0] = y0i[0]*ft_Instrum[0];	/* 0th & 1st elements are separately treated */
	y0i[1] = y0i[1]*ft_Instrum[1];
	for (i=1; i<n/2; i++) {	/* Note i = 1,...,(n/2-1) ! */
		Cmul(&y0i[2*i],&y0i[2*i+1],
			y0i[2*i],y0i[2*i+1],ft_Instrum[2*i],ft_Instrum[2*i+1]);
			/* Multiplication by Fourier transformed instrumental function */
	}
	ifft(y0i,n);	/* inverse Fourier transform of y0i */

	/*** Convolution of error data (1) ***/
	ifft(ft_Instrum,n);	/* now, ft_Instrum is instrumental function */
	for (i=0; i<n; i++) {
		ft_Instrum[i] = sqr(ft_Instrum[i]);
	}	/* now, ft_Instrum is squared instrumental function */
	fft(ft_Instrum,n);
	/* now, ft_Instrum is Fourier transform of squared instrumental function */
	/*** Calculate variance ***/
	for (i=0; i<n; i++) {
		s0[i] = sqr(s0[i]);
	}	/* now, s0 is variance */
	fft(s0,n);
	s0[0] = s0[0] * ft_Instrum[0];
	s0[1] = s0[1] * ft_Instrum[1];
	for (i=1; i<n/2;i++) {
		Cmul(&s0[2*i],&s0[2*i+1],
			s0[2*i],s0[2*i+1],ft_Instrum[2*i],-ft_Instrum[2*i+1]);
		/* Multiplication of conjugate Fourier transform of */
		/* squared instrumental function */
	}
	ifft(s0,n);
	/*** Calculate error ***/
	for (i=0; i<n; i++) {
		if (s0[i] == 0.0) {
			s0[i] = 0.0;
		} else {
			s0[i] = sqrt(fabs(s0[i]));
		}
	}

	if (EvaluateAnotherErrorIndication == 1) {
		/*** Convolution of error data (2) ***/
		ifft(ft_Instrum2,n);	/* now, ft_Instrum is inverse instrumental function */
		for (i=0; i<n; i++) {
			ft_Instrum2[i] = sqr(ft_Instrum2[i]);
		}	/* now, ft_Instrum2 is squared inverse  function */
		fft(ft_Instrum2,n);
			/* now, ft_Instrum2 is Fourier transform of */
			/* squared inverse instrumental function */
		/*** Calculate reciprocal variance ***/
		for (i=0; i<n; i++) {
			if (s1[i] == 0.0) {
				s1[i] = 0.0;
			} else {
				s1[i] = 1/ sqr(s1[i]);
			}
		}	/* now, s1 is reciprocal variance */
		fft(s1,n);
		s1[0] = s1[0] * ft_Instrum2[0];
		s1[1] = s1[1] * ft_Instrum2[1];
		for (i=1; i<n/2;i++) {
			Cmul(&s1[2*i],&s1[2*i+1],
				s1[2*i],s1[2*i+1],ft_Instrum2[2*i],ft_Instrum2[2*i+1]);
				/* Multiplication of Fourier transform of */
				/* squared inverse instrumental function */
		}
		ifft(s1,n);
		/*** Calculate error ***/
		for (i=0; i<n; i++) {
			if (s1[i] == 0.0) {
				s1[i] = 0.0;
			} else {
				s1[i] = 1 / sqrt(fabs(s1[i]));
			}
		}
	}
}

static void PickupData(void)
/* picks up data from interpolated data */
/* uses data "x0T[]", "y0i[]" & "s0[]" */
{
	SplineEq2NonEq(x0start,x0step,y0i,nNonZero,x0T,scat,nData);
	SplineEq2NonEq(x0start,x0step,s0,nNonZero,x0T,err,nData);
	if (EvaluateAnotherErrorIndication == 1) {
		SplineEq2NonEq(x0start,x0step,s1,nNonZero,x0T,err2,nData);
	}
}

static void CreateScaledData(int flag)
/* creates scale transformed data */
/* uses "twot[]", overwrites data "scat[]" & "err[]" */
{
	unsigned long i;
	for (i=0; i<nData; i++) {
		scat[i] *= SpecificWidth(flag,twot[i]);
		scat[i] /= CorrectionFactor(twot[i]);
		err[i] *= SpecificWidth(flag,twot[i]);
		err[i] /= CorrectionFactor(twot[i]);
		if (EvaluateAnotherErrorIndication==1) {
			err2[i] *= SpecificWidth(flag,twot[i]);
			err2[i] /= CorrectionFactor(twot[i]);
		}
	}
#ifdef TRY_EXTRAPOLATION
	scat[nData]=0.5*(scat[nData-1]+scat[0]);//scat[0];
	scat[nData+1]=scat[0];//scat[1];
	err[nData]=0.5*(err[nData-1]+err[0]);//err[0];
	err[nData+1]=err[0];//err[1];
	if (EvaluateAnotherErrorIndication==1) {
		err2[nData]=0.5*(err2[nData-1]+err2[0]);//err2[0];
		err2[nData+1]=err[0];//err2[1];
	}
#endif;
	/*** Pre-correct interpolated error ***/
	PreCorrectInterpolatedError(flag);
}

static void CreateReScaledData(int flag)
/* creates re-scaled data */
/* uses "twot[]", overwrites data "scat[]" & "err[]" */
{
	unsigned long i;

	/*** Re-correct interpolated error ***/
	ReCorrectInterpolatedError(flag);

	for (i=0; i<nData; i++) {
		scat[i] /= SpecificWidth(flag,twot[i]);
		scat[i] *= CorrectionFactor(twot[i]);
		err[i] /= SpecificWidth(flag,twot[i]);
		err[i] *= CorrectionFactor(twot[i]);
		if (EvaluateAnotherErrorIndication==1) {
			err2[i] /= SpecificWidth(flag,twot[i]);
			err2[i] *= CorrectionFactor(twot[i]);
		}
	}
#ifdef TRY_EXTRAPOLATION
	scat[nData]=0.5*(scat[nData-1]+scat[0]);//scat[0];
	scat[nData+1]=scat[0];//scat[1];
	err[nData]=0.5*(err[nData-1]+err[0]);//err[0];
	err[nData+1]=err[0];//err[1];
	if (EvaluateAnotherErrorIndication==1) {
		err2[nData]=0.5*(err2[nData-1]+err2[0]);//err2[0];
		err2[nData+1]=err2[0];//err2[1];
	}
#endif;
}

static void PreCorrectInterpolatedError(int flag)
/* Pre-correction for interpolated error data */
/* uses "twot[]","nData","x0step", overwrites "err[]" */
{
	unsigned long i;
	double x0Tstep;

	for (i=0; i<nData; i++) {
		if (i==0) {
			x0Tstep = x0T[1]-x0T[0];
		} else if (i==nData-1) {
			x0Tstep = x0T[nData-1]-x0T[nData-2];
		} else {
			x0Tstep = (x0T[i+1]-x0T[i-1])/2;
		}

		err[i] *= sqrt(x0Tstep/x0step);
		if (EvaluateAnotherErrorIndication==1) {
			err2[i] *= sqrt(x0Tstep/x0step);
		}
	}
#ifdef TRY_EXTRAPOLATION
	err[nData] *= sqrt(0.5*(x0T[nData+1]-x0T[nData-1])/x0step);
	err[nData+1] *= sqrt((x0T[1]-x0T[0])/x0step);
	if (EvaluateAnotherErrorIndication==1) {
		err2[nData] *= sqrt(0.5*(x0T[nData+1]-x0T[nData-1])/x0step);
		err2[nData+1] *= sqrt((x0T[1]-x0T[0])/x0step);
	}
#endif	
}
static void ReCorrectInterpolatedError(int flag)
/* Post-correction for interpolated error data */
/* uses "twot[]","nData","x0step", overwrites "err[]" */
{
	unsigned long i;
	double x0Tstep;
	for (i=0; i<nData; i++) {
		if (i==0) {
			x0Tstep = x0T[1]-x0T[0];
		} else if (i==nData-1) {
			x0Tstep = x0T[nData-1]-x0T[nData-2];
		} else {
			x0Tstep = (x0T[i+1]-x0T[i-1])/2;
		}
		err[i] /= sqrt(x0Tstep/x0step);
		if (EvaluateAnotherErrorIndication==1) {
			err2[i] /= sqrt(x0Tstep/x0step);
		}
	}
#ifdef TRY_EXTRAPOLATION
	err[nData] /= sqrt(0.5*(x0T[nData+1]-x0T[nData-1])/x0step);
	err[nData+1] /= sqrt((x0T[1]-x0T[0])/x0step);
	if (EvaluateAnotherErrorIndication==1) {
		err2[nData] /= sqrt(0.5*(x0T[nData+1]-x0T[nData-1])/x0step);
		err2[nData+1] /= sqrt((x0T[1]-x0T[0])/x0step);
	}
#endif	
}

void RemoveInstrumental(int flag)
/* removes an instrumental effect */
/* uses data "twot[]", "scat[]" & "err[]" */
{
	/*** Prepare parameters for interpolation ***/
	PrepareParameters(flag);
	/*** Scale transformation ***/
	CreateScaledData(flag);
	/*** Interpolation ***/
	InterpolateData();
	/*** Create Fourier transform of instrumental function ***/
	CreateFTInstrumental(flag,nTotal);
	/*** Deconvolute or Convolute Instrumental function ***/
	if (flag != 0) {
            DeconvoluteInstrumental(flag,nTotal);
        } else {
            ConvoluteInstrumental(flag,nTotal);
        }
	/*** Pickup data (reverse interpolation) ***/
	PickupData();
	/*** Reverse scale transformation ***/
	CreateReScaledData(flag);
	/*** Free memory allocation ***/
	ClearData2();
}

void RemoveAllInstrumental(void)
/* removes all instrumental effect */
{
	#ifdef REPORT
	fprintf(stderr,"*** Deconvolution of Kalpha1, Kalpha2 & Higher-Axial-Divergence ***\n");
	fprintf(stderr,"K_alpha_1 relative width = %f\n",instr.g1);
	fprintf(stderr,"fraction of K_alpha_2 intensity = %f\n",instr.rho);
        fprintf(stderr,"axial divergence angle = %f\n",instr.phiA);
	#endif
	if ((instr.g1==0.0)&&(instr.rho==0.0)&&(instr.phiA==0.0)) {
		#ifdef REPORT
		fprintf(stderr,"...skipped...\n");
		#endif
	} else {
		RemoveInstrumental(1);
	}
	#ifdef REPORT
	fprintf(stderr,"*** Deconvolution of Lower-Axial-Divergence ***\n");
        fprintf(stderr,"axial divergence angle = %f\n",instr.phiA);
	#endif
	if (instr.phiA==0.0) {
		#ifdef REPORT
		fprintf(stderr,"...skipped...\n");
		#endif
	} else {
		RemoveInstrumental(2);
	}
	#ifdef REPORT
	fprintf(stderr,"*** Deconvolution of Sample-Transparency ***\n");
        fprintf(stderr,"penetration depth / gonio.radius = %f\n",instr.invmuR);
	#endif
	if (instr.invmuR==0.0) {
		#ifdef REPORT
		fprintf(stderr,"...skipped...\n");
		#endif
	} else {
		RemoveInstrumental(3);
	}
	#ifdef REPORT
	fprintf(stderr,"*** Deconvolution of Flat-Specimen ***\n");
        fprintf(stderr,"equatorial angle = %f\n",instr.phiE);
        fprintf(stderr,"sample width / gonio.radius = %f\n",instr.W2R);
        fprintf(stderr,"critical diffraction angle = %f\n",instr.twotC);
	#endif
	if ((instr.phiE==0.0)||(instr.W2R==0.0)) {
		#ifdef REPORT
		fprintf(stderr,"...skipped...\n");
		#endif
	} else {
		RemoveInstrumental(4);
	}
	#ifdef REPORT
	fprintf(stderr,"*** Deconvolution of Source X-ray & Receiving Slit Width ***\n");
	fprintf(stderr,"source X-ray relative width = %f\n",instr.S2R);
	fprintf(stderr,"receiving slit relative width = %f\n",instr.R2R);
	#endif
	if ((instr.S2R==0.0)&&(instr.R2R==0.0)) {
		#ifdef REPORT
		fprintf(stderr,"...skipped...\n");
		#endif
	} else {
		RemoveInstrumental(5);
	}
	#ifdef REPORT
//	fprintf(stderr,"*** Convolution of Gaussian Width of %10.5f ***\n",instr.gGmx*2*sqrt(log(2)));
	#endif
//	RemoveInstrumental(0);
}

static void ClearData2(void)
/* free memory allocation */
{
	if (EvaluateAnotherErrorIndication==1) {
		if (ft_Instrum2 != NULL) {
			free(ft_Instrum2);
			ft_Instrum2 = NULL;
		}
	}
	if (ft_Instrum != NULL) {
		free(ft_Instrum);		/* Fourier transform of instrumental function */
		ft_Instrum = NULL;
	}
	if (EvaluateAnotherErrorIndication==1) {
		if (s1!=NULL) {
			free(s1);
			s1 = NULL;
		}
	}
	if (s0!=NULL) {
		free(s0);	/* interpolated error data */
		s0 = NULL;
	}
	if (y0i!=NULL) {
		free(y0i);	/* interpolated intensity data */
		y0i = NULL;
	}
	if (x0T!=NULL) {
		free(x0T);	/* scale transformed abscissa */
		x0T = NULL;
	}
}
