/**************/
/* "spline.c" */
/**************/

/***** INTERFACE *****/

/*** Function Prototype (uses) ***/
void error(char error_text[]);	/* in "error.c" */

/*** Function Prototypes (public) ***/

void SplineNonEq2Eq(double xS[], double yS[], unsigned long nS, 
	double xStart, double xStep, double yD[], unsigned long nD);
	/* in "spline.c" */
	/* Cubic spline interpolation from non-equidistantly sampled data */
	/* to equidistantly sampled data */
void SplineEq2NonEq(double xStart, double xStep, double yS[], unsigned long nS,
	double xD[], double yD[], unsigned long nD);
	/* in "spline.c" */
	/* Cubic spline interpolation from equidistantly sampled data */
	/* to non-equidistantly sampled data */


/***** IMPLEMENTATION *****/

#include <stdlib.h>

/*** Function Prototypes (private) ***/
static double *dvector(unsigned long nl, unsigned long nh);
	/* allocate a double vector with subscript range v[nl..nh] */
static void free_dvector(double *v, unsigned long nl);
	/* free a double vector allocated with dvector() */
static void spline(double x[], double y[], unsigned long n, double yp1, double ypn, 
	double y2[]);	/* in "spline.c" */
	/* Given arrays x[1..n] and y[1..n] containing a tabulated function, */
	/* i.e., y_i = f(x_i), with x_1 < x_2 < ... < x_N , and given values */
	/* yp1 and ypn for the first derivative of the interpolating */
	/* function at points 1 and n, respectively, this routine returns an */
	/* array y2[1..n] that contains the second derivatives of the */
	/* interpolating function at the tabulated points x_i. If yp1 and/or */
	/* ypn are equal to 1e30 or larger, the routine is signaled to set */
	/* the corresponding boundary condition for a natural spline, with */
	/* zero second derivative on that boundary. */
static void splint(double xa[], double ya[], double y2a[], unsigned long n, double x, 
	double *y);	/* in "spline.c" */
	/* Given the arrays xa[1..n] and ya[1..n], which tabulate a function */
	/* (with the xai's in order), and given the array y2a[1..n], which is */
	/* the output from spline above, and given a value of x, this routine */
	/* returns a cubic-spline interpolated value y. */

static double *dvector(unsigned long nl, unsigned long nh)
{
	double *v;

	v=(double *)malloc((size_t) ((nh-nl+1)*sizeof(double)));
	if (!v) error("allocation failure in dvector()");
	return v-nl;
}

static void free_dvector(double *v, unsigned long nl)
{
	free(v+nl);
}

void SplineNonEq2Eq(double xS[], double yS[], unsigned long nS, 
	double xStart, double xStep, double yD[], unsigned long nD)
	/* Cubic spline interpolation from non-equidistantly sampled data */
	/* to equidistantly sampled data */
{
	unsigned long i;
	double *y2;
	
	y2 = malloc(nD*sizeof(double));
	if (y2==NULL) {
		error("Error in memory allocation of y2 in SplineNonEq2Eq().");
		return;
	}
	spline(xS-1,yS-1,nS,1e30,1e30,y2-1);
	for (i=0; i<nD; i++) {
		splint(xS-1,yS-1,y2-1,nS,xStart+i*xStep,&(yD[i]));
	}
	free(y2);
}
void SplineEq2NonEq(double xStart, double xStep, double yS[], unsigned long nS,
	double xD[], double yD[], unsigned long nD)
	/* Cubic spline interpolation from equidistantly sampled data */
	/* to non-equidistantly sampled data */
{
	unsigned long i;
	double *xS, *y2;
	
	xS = malloc(nS*sizeof(double));
	if (xS==NULL) {
		error("Error in memory allocation of xS in SplineEq2NonEq().");
	}
	y2 = malloc(nS*sizeof(double));
	if (y2==NULL) {
		error("Error in memory allocation of y2 in SplineEq2NonEq().");
	}
	for (i=0; i<nS; i++) {
		xS[i] = xStart + i * xStep;
	}
	spline(xS-1,yS-1,nS,1e30,1e30,y2-1);
	for (i=0; i<nD; i++) {
		splint(xS-1,yS-1,y2-1,nS,xD[i],&(yD[i]));
	}
	free(y2);
	free(xS);
}

static void spline(double x[], double y[], unsigned long n, double yp1, double ypn, 
	double y2[])
{
	unsigned long i,k;
	double p,qn,sig,un,*u;
	u=(double *)malloc(n*sizeof(double)); /* instead of dvector(1,n-1); */
	if (u==NULL) 
		error("Could not allocate memory in spline() ");
	if (yp1 > 0.99e30) 
		/* The lower boundary condition is set either to be "natural" */
		y2[1]=u[1]=0.0;
	else { /* or else to have a specified first derivative. */
		y2[1] = -0.5;
		u[1]=(3.0/(x[2]-x[1]))*((y[2]-y[1])/(x[2]-x[1])-yp1);
	}
	for (i=2;i<=n-1;i++) { 
		/* This is the decomposition loop of the tridiagonal */
		/* algorithm. y2 and u are used for temporary storage */
		/* of the decomposed factors. */
		sig=(x[i]-x[i-1])/(x[i+1]-x[i-1]);
		p=sig*y2[i-1]+2.0;
		y2[i]=(sig-1.0)/p;
		u[i]=(y[i+1]-y[i])/(x[i+1]-x[i]) - (y[i]-y[i-1])/(x[i]-x[i-1]);
		u[i]=(6.0*u[i]/(x[i+1]-x[i-1])-sig*u[i-1])/p;
	}
	if (ypn > 0.99e30) 
		/* The upper boundary condition is set either to be */
		/* "natural" */
		qn=un=0.0;
	else { /* or else to have a specified first derivative. */
		qn=0.5;
		un=(3.0/(x[n]-x[n-1]))*(ypn-(y[n]-y[n-1])/(x[n]-x[n-1]));
	}
	y2[n]=(un-qn*u[n-1])/(qn*y2[n-1]+1.0);
	for (k=n-1;k>=1;k--) 
		/* This is the backsubstitution loop of the tridiagonal */
		/* algorithm. */
		y2[k]=y2[k]*y2[k+1]+u[k];
	free(u);	/* instead of free_dvector(u,1); */
}

static void splint(double xa[], double ya[], double y2a[], unsigned long n, double x, double *y)
{
	unsigned long klo,khi,k;
	double h,b,a;
	klo=1; 
		/* We will find the right place in the table by means of */
		/* bisection. This is optimal if sequential calls to this */
		/* routine are at random values of x. If sequential calls */
		/* are in order, and closely spaced, one would do better */
		/* to store previous values of klo and khi and test if */
		/* they remain appropriate on the next call. */
	khi=n;
	while (khi-klo > 1) {
		k=(khi+klo) >> 1;
		if (xa[k] > x) khi=k;
		else klo=k;
	} /* klo and khi now bracket the input value of x. */
	h=xa[khi]-xa[klo];
	if (h == 0.0) error("Bad xa input to routine splint"); 
		/* The xa's must be distinct. */
	a=(xa[khi]-x)/h;
	b=(x-xa[klo])/h; /* Cubic spline polynomial is now evaluated. */
	*y=a*ya[klo]+b*ya[khi]+((a*a*a-a)*y2a[klo]
		+(b*b*b-b)*y2a[khi])*(h*h)/6.0;
}
